#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      units.py
#  brief:     defines units systems
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


import copy
import numpy
from pNbody import iofunc as pnio
from pNbody import param
from sys import exit



##########################################################################

from astropy import units as u


class unitSystem():
  '''
  New system units, compatible with astropy.
  By default the units are the Gadget-2 ones.
  '''
    
  def __init__(self,ULength=3.085678e21*u.cm,UMass=1.989e43*u.g,UVelocity=1e5*u.cm/u.s):
    
    self.ULength   = ULength
    self.UMass     = UMass
    self.UVelocity = UVelocity 
    self.Init()
  
  
  def setUnitLength(self,ULength):
    self.ULength   = ULength
    self.Init()

  def setUnitMass(self,UMass):
    self.UMass   = UMass
    self.Init()
    
  def setUnitVelocity(self,UVelocity):
    self.UVelocity   = UVelocity
    self.Init()        
    
    
  def Init(self):
    
    self.UnitLength               = u.def_unit(['Ul'],self.ULength)     
    self.UnitMass                 = u.def_unit(['Um'],self.UMass)           
    self.UnitVelocity             = u.def_unit(['Uv'],self.UVelocity)  
        
    self.UnitTime                 = u.def_unit(['Ut'], self.UnitLength/self.UnitVelocity)
    self.UnitTemperature          = u.def_unit(['UT'], u.K)
    self.UnitEnergySpec           = self.UnitVelocity**2
    self.UnitEnergy               = self.UnitVelocity**2*self.UnitMass
    self.UnitDensity              = self.UnitMass/self.UnitLength**3
    self.UnitPressure             = self.UnitEnergySpec*self.UnitDensity
    self.UnitBases                = self.UnitLength.bases + self.UnitMass.bases + self.UnitTime.bases + self.UnitTemperature.bases
    



##########################################################################
class Units:
    ##########################################################################
    """
    Units class.
    """

    def __init__(self, symbol, factor=1., power=1, ulist=None):
        """
        """
        self.symbol = symbol
        self.factor = float(factor)
        self.power = int(power)
        if ulist is None:
            self.ulist = []
        else:
            self.ulist = ulist
        self.bases = []

        self.GatherBaseUnits()

    def set_symbol(self, symbol):

        self.ulist = [copy.deepcopy(self)]
        self.power = 1
        self.factor = 1.0
        self.symbol = symbol
        self.GatherBaseUnits()

    def GatherBaseUnits(self):
        """
        create the self.bases list (recursively)
        """

        self.bases = []

        for unit in self.ulist:

            if len(unit.ulist) == 0:		# a base unit
                self.bases.append(unit)   	# add the unit

            else:				# go deeper in the tree
                unit.GatherBaseUnits()
                self.bases = self.bases + unit.bases

        self.ReduceBaseUnits()

    def ReduceBaseUnits(self):

        allunits = {}
        factorTot = 1.0
        for base in self.bases:
            factorTot = factorTot * base.factor
            if base.symbol in allunits:
                allunits[base.symbol] = allunits[base.symbol] + base.power
            else:
                allunits[base.symbol] = base.power

        self.bases = []
        flag = 1
        for key in list(allunits.keys()):
            factor = 1.0
            if flag:
                factor = factorTot
                flag = 0
            power = allunits[key]
            symbol = key
            ulist = []
            self.bases.append(Units(symbol, factor, power, ulist))

    def get_basefactor(self):
        """
        return the factor relative to the base units

        actually, it is stored as the factor of the first base unit
        """
        return self.bases[0].factor

    # def to_units(self,newunits):
    #   """
    #   return a new unit object in new units
    #   """
    #
    #   # check that the units are similar

    def str_base(self):

        string = ''

        if self.factor != 1.0:
            string = string + self.factor.__str__()

        if len(string) > 0:
            string = string + ' '
        if self.symbol is None:
            string = string + '-'
        else:
            string = string + self.symbol

        if self.power != 1:
            string = string + '^%d' % self.power

        return string

    def __str__(self):

        string = ''
        # (1) display using symbol
        if self.factor != 0:
            string = string + self.str_base()

        str1 = "[%s]" % string

        string = ''
        # (2) display using (base units)
        for base in self.bases:
            if len(string) > 0:
                string = string + ' '
            string = string + base.str_base()

        str2 = "[%s]" % string

        if str2 != '[]':
            str1 = "%s\t%s" % (str1, str2)

        return str1

    def __mul__(self, y):

        x = copy.deepcopy(self)
        y = copy.deepcopy(y)

        # multiply with a scalar
        if isinstance(y, int) or isinstance(y, float):
            x.factor = x.factor * y
            x.mul_list(x.ulist, y)
            x.GatherBaseUnits()
            return x

        # multiply with a unit
        elif isinstance(y, type(x)):

            if x.symbol == y.symbol:		# same units
                x.symbol = x.symbol
                x.power = x.power + y.power
                x.ulist = x.ulist + y.ulist
                x.GatherBaseUnits()
                return x

            else:				# different unit
                symbol = None
                power = 1
                factor = 1.0
                if x.ulist == []:			# case of a base unit
                    x.ulist = [copy.deepcopy(x)]
                if y.ulist == []:			# case of a base unit
                    y.ulist = [copy.deepcopy(y)]
                ulist = x.ulist + y.ulist
                return Units(symbol, factor, power, ulist)

    def __truediv__(self, y):

        x = copy.deepcopy(self)
        y = copy.deepcopy(y)

        # divide with a scalar
        if isinstance(y, int) or isinstance(y, float):
            x.factor = x.factor / y
            x.div_list(x.ulist, y)
            x.GatherBaseUnits()
            return x

        # divide with a unit
        elif isinstance(y, type(x)):

            if x.symbol == y.symbol:		# same units
                x.symbol = None
                x.power = 0
                x.ulist = []
                x.GatherBaseUnits()
                return x

            else:				# different unit
                symbol = None
                power = 1
                factor = 1.0
                if x.ulist == []:			# case of a base unit
                    x.ulist = [copy.deepcopy(x)]
                if y.ulist == []:			# case of a base unit
                    y.ulist = [copy.deepcopy(y)]
                y.pow_list(y.ulist, -1)
                y.GatherBaseUnits()
                ulist = x.ulist + y.ulist
                return Units(symbol, factor, power, ulist)

    def __rtruediv__(self, y):
        x = copy.deepcopy(self)

        # divide with a scalar
        if isinstance(y, int) or isinstance(y, float):
            return y * x**-1
        else:
            raise Exception(
                "TypeError",
                "unsupported operand type(s) for / or div()")

    def __pow__(self, y):

        x = copy.deepcopy(self)
        y = copy.deepcopy(y)

        # power with a scalar
        if isinstance(y, int) or isinstance(y, float):
            x.factor = x.factor**y
            x.power = x.power * y
            x.pow_list(x.ulist, y)
            x.GatherBaseUnits()
            return x

        else:
            raise Exception(
                "TypeError",
                "unsupported operand type(s) for ** or pow()")

    def mul_list(self, ulist, y):

        for u in ulist:
            if len(u.ulist) == 0:		# a base unit
                u.factor = u.factor * y
            else:				# go deeper in the tree
                u.mul_list(u.ulist, y)
                u.GatherBaseUnits()

    def div_list(self, ulist, y):

        for u in ulist:
            if len(u.ulist) == 0:		# a base unit
                u.factor = u.factor / y
            else:				# go deeper in the tree
                u.mul_list(u.ulist, y)
                u.GatherBaseUnits()

    def pow_list(self, ulist, y):

        for u in ulist:
            if len(u.ulist) == 0:		# a base unit
                u.factor = u.factor**y
                u.power = u.power * y
            else:				# go deeper in the tree
                u.pow_list(u.ulist, y)
                u.GatherBaseUnits()

    __rmul__ = __mul__


##########################################################################
class UnitSystem:
    ##########################################################################
    """
    Units system
    """

    def __init__(self, UnitSysName, UnitLst):
        """
        UnitDic = {'length':UnitLength,'mass':UnitMass,'time':UnitTime}

        The units used to define a system of units must be "basic", i.e.,
        not based on several units


        useful variables

        self.dic_of_factors		: value of base unit in current units
        self.UnitDic		    : contains base units and value of the chosen unit in this base units

        """

        UnitDic = {}
        for u in UnitLst:
            u.GatherBaseUnits()

            if len(u.bases) > 1:
                raise Exception(
                    "UnitSystemError",
                    "%s is not a pure Unit." %
                    u.symbol)

            elif len(u.bases) == 1:
                UnitDic[u.bases[0].symbol] = u.bases[0]

            elif len(u.bases) == 0:				# base unit
                UnitDic[u.symbol] = u

        self.set_dics(UnitDic)

        self.UnitSysName = UnitSysName
        self.UnitLst = UnitLst

    def set_dics(self, UnitDic):

        dic_of_factors = {}
        dic_of_powers = {}

        for UnitType in list(UnitDic.keys()):
            u = UnitDic[UnitType]

            if u.bases == []:					# base unit
                dic_of_factors[UnitType] = 1 / u.factor
                dic_of_powers[UnitType] = u.power
            else:
                dic_of_factors[UnitType] = 1 / u.bases[0].factor
                dic_of_powers[UnitType] = u.bases[0].power

        self.dic_of_factors = dic_of_factors		# value of base unit in current units
        self.dic_of_powers = dic_of_powers			# not useful

        self.UnitDic = UnitDic

        self.UnitLength = self.UnitDic['m']
        self.UnitMass = self.UnitDic['kg']
        self.UnitTime = self.UnitDic['s']

        self.UnitVelocity = self.UnitDic['m'] / self.UnitDic['s']
        self.UnitSpecEnergy = self.UnitVelocity**2
        self.UnitEnergy = self.UnitSpecEnergy * self.UnitMass

        self.UnitDensity = self.UnitMass / self.UnitLength**3
        self.UnitSurfaceDensity = self.UnitMass / self.UnitLength**2
        self.UnitSurface = self.UnitLength**2

    def get_UnitLength_in_cm(self):
        f = self.UnitDic['m'].factor
        c = PhysCte(f, Unit_m)
        return c.into(cgs)

    def get_UnitMass_in_g(self):
        f = self.UnitDic['kg'].factor
        c = PhysCte(f, Unit_kg)
        return c.into(cgs)

    def get_UnitTime_in_s(self):
        f = self.UnitDic['s'].factor
        c = PhysCte(f, Unit_s)
        return c.into(cgs)

    def get_UnitVelocity_in_cm_per_s(self):
        return self.get_UnitLength_in_cm() / self.get_UnitTime_in_s()

    def get_UnitEnergy_in_cgs(self):
        return self.get_UnitMass_in_g() * self.get_UnitVelocity_in_cm_per_s()**2

    def get_UnitDensity_in_cgs(self):
        return self.get_UnitMass_in_g() / self.get_UnitLength_in_cm()**3

    def info(self):
        """
        print some info
        """
        print("units info")
        print(("  UnitLength_in_cm         =%g" %
               (self.get_UnitLength_in_cm())))
        print(("  UnitVelocity_in_cm_per_s =%g" %
               (self.get_UnitVelocity_in_cm_per_s())))
        print(("  UnitMass_in_g            =%g" % (self.get_UnitMass_in_g())))

    def getparam(self):
        param = {}
        param["UnitLength_in_cm"] = self.get_UnitLength_in_cm()
        param["UnitVelocity_in_cm_per_s"] = self.get_UnitVelocity_in_cm_per_s()
        param["UnitMass_in_g"] = self.get_UnitMass_in_g()
        return param

    def convertionFactorTo(self, newUnits):
        """
        return the conversion factor to obtain the new units
        """

        f = 1.0

        # loop over all base units of the new Units
        if newUnits.bases == []:
            baselist = [newUnits]
        else:
            baselist = newUnits.bases

        for base in baselist:

            factor = base.factor
            symbol = base.symbol
            power = base.power

            # multiply by the right factor
            f = f / (factor * (self.dic_of_factors[symbol])**power)

        return f

    def into(self, newUnits):
        """
        return into the new units
        """
        return PhysCte(self.convertionFactorTo(newUnits), newUnits)

    def multiply_UnitLength(self, f):
        self.UnitDic['m'].factor *= f
        self.set_dics(self.UnitDic)

    def multiply_UnitMass(self, f):
        self.UnitDic['kg'].factor *= f
        self.set_dics(self.UnitDic)

    def multiply_UnitTime(self, f):
        self.UnitDic['s'].factor *= f
        self.set_dics(self.UnitDic)

    def CorrectFromHubbleParameter(self, HubbleParam):
        print(("Units Correction : HubbleParam = %g" % HubbleParam))
        self.multiply_UnitLength(1.0 / HubbleParam)
        self.multiply_UnitMass(1.0 / HubbleParam)
        self.multiply_UnitTime(1.0 / HubbleParam)


    def to_unitSystem(self):
      ULength = self.get_UnitLength_in_cm()*u.cm
      UMass = self.get_UnitMass_in_g()*u.g
      UVelocity = self.get_UnitVelocity_in_cm_per_s()*u.cm/u.s
    
      return unitSystem(ULength=ULength,UMass=UMass,UVelocity=UVelocity)




##########################################################################
class PhysCte():
    ##########################################################################
    """
    Physical constant
    """

    def __init__(self, value, Unit):
        # super(PhysCte,self).__init__(value)
        self.value = value
        self.Unit = Unit
        if isinstance(Unit, (list, numpy.ndarray)):
            msg = "Can't handle non-scalars with varying units. All units need to be the same."
            raise ValueError(msg)

    def factor_to_base(self):

        if self.Unit.bases == []:
            factor = self.Unit.factor
        else:
            factor = self.Unit.bases[0].factor

        return factor

    def __str__(self):
        if isinstance(self.value, (list, numpy.ndarray)):
            return self.value.__str__() + ' [%s]' % self.Unit
           # return self.value.dtype.__str__() + ' %s' % self.Unit
        else:
            return self.value.__str__() + ' [%s]' % self.Unit

    def into(self, SystemName):			# here, we could return an object

        f = self.value

        if self.Unit.bases == []:
            unit = self.Unit
            symbol = unit.symbol
            f = f * unit.factor * \
                (SystemName.dic_of_factors[symbol])**unit.power
        else:
            for unit in self.Unit.bases:
                symbol = unit.symbol
                f = f * unit.factor * \
                    (SystemName.dic_of_factors[symbol])**unit.power

        return float(f)

    def __float__(self):
        if isinstance(self.value, numpy.ndarray):
            print(self[0])
            return(self[:].value.astype(numpy.float))
        else:
            return float(self.value)

    def __mul__(self, y):
        if hasattr(y, 'Unit'):
            # if multiplying with something that has a unit, i.e. another PhysCte
            return PhysCte(self.value * y.value, self.Unit * y.Unit)
        elif isinstance(y, Units):
            from . import message
            message.todo_warning()
            return self
            # this doesn't work
            # return PhysCte(self.value * y.factor, self.Unit*y)
        else:
            return PhysCte(self.value*y, self.Unit)

    __rmul__ = __mul__

    def __add__(self, y):
        pass

    def __sub__(self, y):
        pass

    def __truediv__(self, y):
        if hasattr(y, 'Unit'):
            # if multiplying with something that has a unit
            return PhysCte(self.value / y.value, self.Unit / y.Unit)
        elif isinstance(y, Units):
            from . import message
            message.todo_warning()
            return self
            # this doesn't work
            # return PhysCte(self.value / y.factor, self.Unit/y)
        else:
            return PhysCte(self.value / y, self.Unit)
    __rtruediv__ = __truediv__

    def __pow__(self, y):
        pass


##########################################################################
# define some units
##########################################################################


# base units
Unit_m = Units('m')
Unit_kg = Units('kg')
Unit_s = Units('s')
Unit_mol = Units('mol')
Unit_C = Units('C')
Unit_K = Units('K')


# other length units
Unit_cm = 0.01 * Unit_m
Unit_cm.set_symbol('cm')
Unit_km = 1e3 * Unit_m
Unit_km.set_symbol('km')
Unit_Mm = 1e6 * Unit_m
Unit_Mm.set_symbol('Mm')
Unit_Gm = 1e9 * Unit_m
Unit_Gm.set_symbol('Gm')
Unit_kpc = 1000 * 3.085e18 * Unit_cm
Unit_kpc.set_symbol('kpc')
Unit_Mpc = 1000000 * 3.085e18 * Unit_cm
Unit_Mpc.set_symbol('Mpc')
Unit_pc = 3.085e18 * Unit_cm
Unit_pc.set_symbol('pc')
Unit_ua = 1.495978e11 * Unit_m
Unit_ua.set_symbol('ua')

# other mass units
Unit_g = 0.001 * Unit_kg
Unit_g.set_symbol('g')
Unit_Ms = 1.9891e33 * Unit_g
Unit_Ms.set_symbol('Ms')
Unit_Msol = 1.9891e33 * Unit_g
Unit_Msol.set_symbol('Ms')
Unit_Mg = 2.23e11 * Unit_Ms
Unit_Mg.set_symbol('Mg')
Unit_Mt = 5.9742e24 * Unit_kg
Unit_Mt.set_symbol('Mt')
Unit_Mj = 317.893 * Unit_Mt
Unit_Mj.set_symbol('Mj')

# other time units
Unit_h = 3600 * Unit_s
Unit_h.set_symbol('h')
Unit_yr = 31536000 * Unit_s
Unit_yr.set_symbol('yr')
Unit_kyr = 1e3 * Unit_yr
Unit_kyr.set_symbol('kyr')
Unit_Myr = 1e6 * Unit_yr
Unit_Myr.set_symbol('Myr')
Unit_Gyr = 1e9 * Unit_yr
Unit_Gyr.set_symbol('Gyr')
Unit_dy = 86400 * Unit_s
Unit_dy.set_symbol('days')
Unit_hr = 3600 * Unit_s
Unit_hr.set_symbol('hr')
Unit_century = 100 * Unit_yr
Unit_century.set_symbol('century')

# other speed units
Unit_kmh = Unit_km / Unit_h
Unit_kmh.set_symbol('kmh')
Unit_kms = Unit_km / Unit_s
Unit_kms.set_symbol('kms')

# other units
Unit_ms = Unit_m / Unit_s
Unit_ms2 = Unit_ms**2
Unit_J = Unit_kg * Unit_ms2
Unit_J.set_symbol('J')
Unit_erg = Unit_g * (Unit_cm / Unit_s)**2
Unit_erg.set_symbol('erg')
Unit_N = Unit_kg * Unit_m / Unit_s**2
Unit_N.set_symbol('N')
Unit_Pa = Unit_N / Unit_m**2
Unit_Pa.set_symbol('Pa')
Unit_G = Unit_N * Unit_m**2 / Unit_kg**2
Unit_d = Unit_kg / Unit_m**3
Unit_Lsol = 3.839 * Unit_erg / Unit_s
Unit_Lsol.set_symbol('Lsol')

##########################################################################
# define some common unit systems
##########################################################################

mks = UnitSystem('mks', [Unit_m, Unit_kg, Unit_s, Unit_K, Unit_mol, Unit_C])
cgs = UnitSystem('cgs', [Unit_cm, Unit_g, Unit_s, Unit_K, Unit_mol, Unit_C])
gal = UnitSystem(
    'gal', [
        Unit_kpc, Unit_Mg, Unit_Myr, Unit_K, Unit_mol, Unit_C])





##########################################################################
# define some functions
##########################################################################


#################################
def Set_SystemUnits_From_Params(params):
    #################################
    """
    return a system of units from given parameters

    params is a dictionary that must contain at least

        params['UnitVelocity_in_cm_per_s'],

        params['UnitMass_in_g'],

        params['UnitLength_in_cm']

    """
    UnitVelocity_in_cm_per_s = params['UnitVelocity_in_cm_per_s']
    UnitMass_in_g = params['UnitMass_in_g']
    UnitLength_in_cm = params['UnitLength_in_cm']
    UnitTime_in_s = UnitLength_in_cm / UnitVelocity_in_cm_per_s

    # now from the params, define a system of units

    Unit_length = Unit_cm * UnitLength_in_cm
    Unit_mass = Unit_g * UnitMass_in_g
    Unit_time = Unit_s * UnitTime_in_s

    localsystem = UnitSystem(
        'local', [
            Unit_length, Unit_mass, Unit_time, Unit_K, Unit_mol, Unit_C])

    return localsystem


#################################
def Set_SystemUnits_From_File(unitsfile):
    #################################
    """
    return a system of units from given file coding units

    unitsfile file is either a gadget parameter file or a pNbody  units file


    """
    params = {}

    # try to read a gadget file
    try:
        gparams = pnio.read_params(unitsfile)
        params['UnitLength_in_cm'] = gparams['UnitLength_in_cm']
        params['UnitMass_in_g'] = gparams['UnitMass_in_g']
        params['UnitVelocity_in_cm_per_s'] = gparams['UnitVelocity_in_cm_per_s']

        # params['Omega0']                   = gparams['Omega0']
        # params['OmegaLambda']              = gparams['OmegaLambda']
        # params['OmegaBaryon']              = gparams['OmegaBaryon']
        # params['BoxSize']                  = gparams['BoxSize']
        # params['ComovingIntegrationOn']    = gparams['ComovingIntegrationOn']

        params = gparams

    except BaseException:

        # try to read a pNbody units file
        try:
            gparams = param.Params(unitsfile, None)

            params['UnitLength_in_cm'] = gparams.get("UnitLength_in_cm")
            params['UnitMass_in_g'] = gparams.get("UnitMass_in_g")
            params['UnitVelocity_in_cm_per_s'] = gparams.get(
                "UnitVelocity_in_cm_per_s")

            # params['Omega0']                   = gparams.get('Omega0')
            # params['OmegaLambda']              = gparams.get('OmegaLambda')
            # params['OmegaBaryon']              = gparams.get('OmegaBaryon')
            # params['BoxSize']                  = gparams.get('BoxSize')
            # params['ComovingIntegrationOn']    = gparams.get('ComovingIntegrationOn')

        except BaseException:
            raise IOError(
                0o15,
                'format of unitsfile %s unknown ! Pease check.' %
                unitsfile)

    localsystem = Set_SystemUnits_From_Params(params)

    return localsystem


def GetUnitsFromString(UnitString):

    from pNbody import ctes

    #################
    # Mass Units
    #################
    if UnitString == "Msol":  # Msol
        return Unit_Msol

    elif UnitString == "g":       # gramm
        return Unit_g

    #################
    # Length Units
    #################
    elif UnitString == "cm":  # cm
        return Unit_cm

    elif UnitString == "pc":  # pc
        return Unit_pc

    elif UnitString == "kpc":  # kpc
        return Unit_kpc

    elif UnitString == "Mpc":  # Mpc
        return Unit_Mpc

    #################
    # Time Units
    #################
    elif UnitString == "yr":  # years
        return Unit_yr

    elif UnitString == "Myr":  # mega years
        return Unit_Myr

    elif UnitString == "Gyr":  # giga years
        return Unit_Gyr

    #################
    # Density Units
    #################
    elif UnitString == "a/cm3" or UnitString == "acc":  # atom/cm^3
        Unit_atom = ctes.PROTONMASS.into(cgs) * Unit_g
        return Unit_atom / (Unit_cm**3)

    elif UnitString == "g/cm3":		# gram/cm^3
        return Unit_g / (Unit_cm**3)

    elif UnitString == "Msol/kpc3":  # Msol/kpc3
        return Unit_Msol / (Unit_kpc**3)

    elif UnitString == "Msol/pc3":  # Msol/pc3
        return Unit_Msol / (Unit_pc**3)

    #################
    # Velocity Units
    #################

    elif UnitString == "m/s":		# m/s
        return Unit_m / Unit_s

    elif UnitString == "km/s":		# m/s
        return Unit_km / Unit_s

    #################
    # Energy Units
    #################

    elif UnitString == "erg":		# gram * (cm/s)**2
        return Unit_g * (Unit_cm**2) / (Unit_s**2)

    else:
        print(UnitString)
        print("unknown unit name")
        exit()



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      observations.py
#  brief:     observations.py
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################

import numpy as np
from pickle import load
from h5py import File, special_dtype
from os import path

# Transformation between pickle host name to HDF5
pickle_to_hdf5 = {
    "MW": "MilkyWay Dwarfs",
    "M31": "Andromeda Dwarfs"
}

# All the required pickle fields
pickle_fields = [
    'eFeHp', 'eVmagp', 'eVmagm', 'Rh', 'eFeHm', 'Vmag',
    'D_max', 'MHI', 'FeH', 'ra', 'eMvp', 'D', 'eLm', 'L',
    'Mv', 'D_min', 'eMvm', 'eLp', 'Sigma',
    'eSigmam', 'eRhp', 'mM0', 'FeH_Tech', 'eRhm', 'eSigmap',
    'dec'
]

# List of fields containing strings
pickle_string_fields = [
    "FeH_Tech",
    "host"
]

# Transformation between pickle fields and HDF5 fields
pickle_fields_to_hdf5 = {
    'Rh': "HalfLightRadius",
    'MHI': "MassHI",
    'FeH': "FeH",
    'ra': "RightAscension",
    'D': "Distance",
    'L': "Luminosity",
    'FeH_Tech': "FeHTechnique",
    'dec': "Declination",
    'Sigma': "VelocityDispersion",
    'host': "Host"
}

# Error fields corresponding to a pickle field
pickle_error_fields = {
    "Rh": ("rel", ["eRhm", "eRhp"]),
    "MHI": None,
    "FeH": ("rel", ["eFeHm", "eFeHp"]),
    "D": ("abs", ["D_min", "D_max"]),
    "L": ("rel", ["eLm", "eLp"]),
    "FeH_Tech": None,
    "dec": None,
    "ra": None,
    "Sigma": ("rel", ["eSigmam", "eSigmap"]),
    "host": None,
}

# Description of each pickle field
pickle_fields_description = {
    "Rh": "Half Light Radius",
    "MHI": "Total mass of HI",
    "FeH": "[Fe/H]",
    "D": "Distance from the Sun",
    "L": "Luminosity",
    "FeH_Tech": "Technique used to measure [Fe/H]",
    "dec": "Declination",
    "ra": "Right Ascension",
    "Sigma": "Stellar velocity dispersion",
    "host": "Host of the satellite [see /Hosts for details]",
}

# Units of each field (in astropy convention)
pickle_fields_units = {
    "FeH": "",
    "D": "kpc",
    "Rh": "kpc",
    "MHI": "solMass",
    "L": "solLum",
    "FeH_Tech": None,
    "dec": "deg",
    "ra": "deg",
    "Sigma": "km/s",
    "host": None,
}

# ASCII format
ascii_data_type = [
    ("name", "S20"),
    ("ra", "f8"),
    ("dec", "f8"),
    ("D", "f8"),
    ("Mass", "f8"),
    ("Radius", "f8"),
    ("source", "S200")
]


class Observations():
    def __init__(self, filename, dataset=None):
        # init fields
        self.filename = filename
        self.data = None
        self.dataset = dataset

        # read data
        self.readHDF5File(filename)

    def readHDF5File(self, filename):
        return

    def useDataset(self, dataset):
        if (self.dataset is None or
                dataset in self.dataset):
            return True

        return False

    def getPosition(self):
        f = File(self.filename)
        x = []
        y = []
        z = []

        for name in f:
            if not self.useDataset(name):
                continue
            # get values
            ra = f[name + "/RightAscension"]
            dec = f[name + "/Declination"]
            # discard errors
            D = f[name + "/Distance"][:, 0]

            # compute position
            x_tmp = D * np.cos(dec) * np.cos(ra)
            y_tmp = D * np.cos(dec) * np.sin(ra)
            z_tmp = D * np.sin(ra)

            x.extend(x_tmp)
            y.extend(y_tmp)
            z.extend(z_tmp)

        pos = np.array([x, y, z]).transpose()
        return pos

    @staticmethod
    def _generateArraysFromPickle(p_data, host):
        """
        Pick all the arrays for a host from the pickle data

        Parameters
        ----------

        p_data: dict
            The whole pickle file data

        host: str
            Name of the host (in HDF5)
        """
        ret = {}

        # initialize data
        for f in pickle_fields:
            ret[f] = []
        ret["name"] = []

        # set data
        for gal_name in p_data:
            gal = p_data[gal_name]
            gal_host = pickle_to_hdf5[gal["host"]]
            if gal_host != host:
                continue

            for f in pickle_fields:
                tmp = gal[f]
                if tmp is None:
                    tmp = np.nan
                ret[f].append(tmp)
            ret["name"].append(gal_name)

        return ret

    @staticmethod
    def _writeArrays(grp, data):
        """
        Write all the arrays to an HDF5 group.

        Parameters
        ----------

        grp: HDF5 group
            The HDF5 group to write

        data: dict
            All the arrays to write
        """
        string_dtype = special_dtype(vlen=str)
        for i in data:
            if i not in list(pickle_fields_to_hdf5.keys()):
                continue

            arr = np.array(data[i])

            err_fields = pickle_error_fields[i]
            if err_fields is not None:
                tpe = err_fields[0]
                fields = err_fields[1]

                arr = arr[:, np.newaxis]
                ret_arr = arr
                # add errors
                for j in fields:
                    tmp = np.array(data[j])[:, np.newaxis]
                    if tpe == "abs":
                        tmp = np.abs(tmp - ret_arr)
                    arr = np.append(arr, tmp, axis=1)

            # write dataset
            name = pickle_fields_to_hdf5[i]
            if i in pickle_string_fields:
                ds = grp.create_dataset(name, data=arr, dtype=string_dtype)
            else:
                ds = grp.create_dataset(name, data=arr)

            # write attributes
            ds.attrs["Description"] = pickle_fields_description[i]
            units = pickle_fields_units[i]
            if units is not None:
                ds.attrs["Units"] = units

        names = np.array(data["name"])
        grp.create_dataset("Name", data=names, dtype=string_dtype)

    @staticmethod
    def transformPickleFile(f_read, f_write, source=None):
        """
        Transform a pickle file containing observations to an HDF5 file.

        You can decide which fields are written and with the arrays at
        the beginning of this document.

        Parameters
        ----------

        f_read: str
            Pickle filename to read

        f_write: str
            HDF5 filename to write
        """

        # open files
        with open(f_read, "rb") as f:
            p_data = load(f)

        f = File(f_write, "a")

        # get list of all hosts
        hosts = []
        for gal in p_data:
            host = p_data[gal]["host"]
            if host not in hosts:
                hosts.append(host)

        # Transform hosts to HDF5 format
        for i, h in enumerate(hosts):
            hosts[i] = pickle_to_hdf5[h]

        # check if only new database
        for h in hosts:
            if h in f:
                raise IOError(
                    "Cannot overwrite an existing group (%s)" % h)

        # create data base
        for h in hosts:
            grp = f.create_group(h)

            # Generate host data
            data = Observations._generateArraysFromPickle(p_data, h)

            # Write the data to HDF5
            Observations._writeArrays(grp, data)

            # Write source
            if source is None:
                source = "Unknown"
            grp.attrs["Source"] = source
            grp.attrs["From file"] = f_read

        # Close file
        f.close()

    @staticmethod
    def transformASCIIFile(f_read, f_write, source=None, group_name=None):
        # read data
        data = np.genfromtxt(f_read, dtype=ascii_data_type, delimiter=",")

        # Open output
        f = File(f_write, "a")

        # create data base
        if group_name is None:
            group_name = path.splitext(f_read)[0]

        grp = f.create_group(group_name)

        # Write data to HDF5



#!/usr/bin/env python3
###########################################################################################
#  package:   pNbody
#  file:      palette.py
#  brief:     Color palette for plots
#  copyright: GPLv3
#             Copyright (C) 2019 EPFL (Ecole Polytechnique Federale de Lausanne)
#             LASTRO - Laboratory of Astrophysics of EPFL
#  author:    Yves Revaz <yves.revaz@epfl.ch>
#
# This file is part of pNbody.
###########################################################################################


"""
this module is used to deal with color palettes.
"""

try:
    import tkinter as tk
    from PIL import ImageTk
    is_tk = True
except ImportError:
    is_tk = False

import numpy as np
import os
import string
import glob


# import parameters
from .parameters import *


# !!! scipy import automatiquement Image qui entre en conflit avec Image
# le mieux est donc de remplacer explicitement Image par Image de scipy
# (qui est le meme d'ailleurs)

try:
    from scipy.misc.pilutil import Image
    from scipy.interpolate import splrep, splev
    is_scipy = True
except ImportError:
    from PIL import Image
    is_scipy = False


##########################################################################
# read lut (palette)
##########################################################################


def readlut(filename=os.path.join(PALETTEDIR, DEFAULTPALETTE)):
    """
    Read a lut file.
    """

    numByte = 256

    file = open(filename, "r")

    data = file.read(numByte)
    data = file.read(numByte)

    # read red
    data = file.read(numByte)
    red = np.fromstring(data, 'b')

    # read green
    data = file.read(numByte)
    green = np.fromstring(data, 'b')

    # read blue
    data = file.read(numByte)
    blue = np.fromstring(data, 'b')

    file.close()

    # combine the colors

    pal = chr(red[0]) + chr(green[0]) + chr(blue[0])

    for i in range(1, 256):
        pal = pal + chr(red[i]) + chr(green[i]) + chr(blue[i])

    return pal


def read_gimp_palette(name):

    f = open(name)
    f.readline()
    pal = ''
    for i in range(255):
        line = f.readline()
        line = str.split(line)
        r = int(line[0])
        g = int(line[1])
        b = int(line[2])

        pal = pal + chr(r) + chr(g) + chr(b)

    return pal

##########################################################################
# Create a palette
##########################################################################


def getpalette(palettename):

    try:
        pal = read_gimp_palette(palettename)
    except BaseException:
        try:
            pal = readlut(palettename)
        except BaseException:
            pal = readlut()

    return(pal)


##########################
class EditParams:
    ##########################

    def __init__(self, master, c, w, d, r):

        self.master = master

        self.root = Toplevel()

        self.frame = Frame(self.root)
        self.frame.grid(column=0, row=0)

        # lablel c
        self.labelc = Label(self.frame, text='valeur c = ')
        self.labelc.grid(column=0, row=0)

        # entry c
        self.entryc = Entry(self.frame)
        self.entryc.grid(column=1, row=0)
        self.entryc.insert(INSERT, "%s" % c)

        # lablel w
        self.labelw = Label(self.frame, text='valeur w = ')
        self.labelw.grid(column=0, row=1)

        # entry w
        self.entryw = Entry(self.frame)
        self.entryw.grid(column=1, row=1)
        self.entryw.insert(INSERT, "%s" % w)

        # lablel d
        self.labeld = Label(self.frame, text='valeur d = ')
        self.labeld.grid(column=0, row=2)

        # entry d
        self.entryd = Entry(self.frame)
        self.entryd.grid(column=1, row=2)
        self.entryd.insert(INSERT, "%s" % d)

        # lablel r
        self.labelr = Label(self.frame, text='valeur r = ')
        self.labelr.grid(column=0, row=3)

        # entry r
        self.entryr = Entry(self.frame)
        self.entryr.grid(column=1, row=3)
        self.entryr.insert(INSERT, "%s" % r)

        # buttons
        self.sendbutton = Button(self.frame, text='send', command=self.send)
        self.sendbutton.grid(column=0, row=4)

        self.okbutton = Button(self.frame, text='ok', command=self.ok)
        self.okbutton.grid(column=1, row=4)

        self.cancelbutton = Button(
            self.frame, text='cancel', command=self.cancel)
        self.cancelbutton.grid(column=2, row=4)

    def ok(self):

        try:
            c = int(self.entryc.get())
        except BaseException:
            print("invalid value for c")

        try:
            w = int(self.entryw.get())
        except BaseException:
            print("invalid value for w")

        try:
            d = float(self.entryd.get())
        except BaseException:
            print("invalid value for d")

        try:
            r = int(self.entryr.get())
        except BaseException:
            print("invalid value for r")

        # send values to main
        self.master.c = c
        self.master.w = w
        self.master.d = d
        self.master.invert = r
        self.master.get(c=c, w=w, d=d, invert=r)
        self.master.draw()

        self.root.destroy()

    def send(self):

        try:
            c = int(self.entryc.get())
        except BaseException:
            print("invalid value for c")

        try:
            w = int(self.entryw.get())
        except BaseException:
            print("invalid value for w")

        try:
            d = float(self.entryd.get())
        except BaseException:
            print("invalid value for d")

        try:
            r = int(self.entryr.get())
        except BaseException:
            print("invalid value for r")

        # send values to main
        self.master.c = c
        self.master.w = w
        self.master.d = d
        self.master.invert = r
        self.master.get(c=c, w=w, d=d, invert=r)
        self.master.draw()

    def cancel(self):
        self.root.destroy()


##########################################################################
# Class Palette
##########################################################################

class Palette:

    def __init__(self, name='light'):

        self.name = self.check_palette_name(name)

        self.Canvas = None

        self.tables = glob.glob(os.path.join(PALETTEDIR, '*'))

        self.c = 128
        self.w = 256
        self.d = 256.
        self.invert = 0

        self.read(self.name)

    ###########################
    def check_palette_name(self, name):
        ###########################

        if os.path.isfile(name):
            pass
        else:
            name = os.path.join(PALETTEDIR, name)

        if not os.path.exists(name):
            print((name, "do not exists, using %s instead" % (DEFAULTPALETTE)))
            name = os.path.join(PALETTEDIR, DEFAULTPALETTE)

        return name

    ################################

    def read(self, name):
        ################################

        t = []
        r = []
        g = []
        b = []

        f = open(name)
        f.readline()
        pal = ''
        for i in range(256):
            line = f.readline()
            line = str.split(line)
            t.append(float(i))
            r.append(float(line[0]))
            g.append(float(line[1]))
            b.append(float(line[2]))

        self.t = np.array(t, float)
        self.r = np.array(r, float)
        self.g = np.array(g, float)
        self.b = np.array(b, float)

        # store initial values
        self.t0 = self.t
        self.r0 = self.r
        self.g0 = self.g
        self.b0 = self.b

        self.mkspline()
        # spline should not be done here
        self.get(c=self.c, w=self.w, d=self.d)

    ################################
    def write(self, name):
        ################################

        f = open(name, 'w')
        f.write("# table rgb (%s)\n" % name)

        for i in range(256):
            f.write("%3d %3d %3d\n" % (self.r[i], self.g[i], self.b[i]))

        f.close()

    ################################
    def mkspline(self):
        ################################

        if is_scipy:
            # splines
            self.ar = splrep(self.t, self.r, s=0)
            self.ag = splrep(self.t, self.g, s=0)
            self.ab = splrep(self.t, self.b, s=0)

    ################################

    def append(self, p):
        ################################
        """
        add a new palette
        """

        c = self.r + self.g + self.b

        self.r = np.where(c == 0, p.r, self.r)
        self.g = np.where(c == 0, p.g, self.g)
        self.b = np.where(c == 0, p.b, self.b)

    ################################
    def getr(self, vals):
        ################################
        """
        return r value
        """
        vals = np.clip(vals, 0, 255)
        return self.r[vals]

    ################################
    def getg(self, vals):
        ################################
        """
        return g value
        """
        vals = np.clip(vals, 0, 255)
        return self.g[vals]

    ################################
    def getb(self, vals):
        ################################
        """
        return b value
        """
        vals = np.clip(vals, 0, 255)
        return self.b[vals]

    ################################
    def get(self, c=0, w=0, d=256., invert=0):
        ################################
        """
        return the palette

        w : width of the interval [0,255]
        c : center of the interval [0,255]

        """

        if w == 0 or c == 0:
            xmin = 0
            xmax = 256
        else:
            xmin = c - w / 2
            xmax = c + w / 2

        self.t = np.arange(xmin, xmax, (xmax - xmin) / 256.)

        # fonction logarithmique pour exemple
        if d < 256.:
            self.t = (np.log((self.t - xmin) / d + 1) /
                      np.log((xmax - xmin) / d + 1)) * (xmax - xmin) + xmin

        if is_scipy:
            self.r = splev(self.t, self.ar)
            self.g = splev(self.t, self.ag)
            self.b = splev(self.t, self.ab)

        self.r = np.clip(self.r, 0., 255.)
        self.g = np.clip(self.g, 0., 255.)
        self.b = np.clip(self.b, 0., 255.)

        # inversion
        if invert:
            self.r = 255. - self.r
            self.g = 255. - self.g
            self.b = 255. - self.b

        self.palette = b''

        for i in range(256):

            r = int(self.r[i])
            g = int(self.g[i])
            b = int(self.b[i])

            self.palette = self.palette + \
                bytearray((r,)) + bytearray((g,)) + bytearray((b,))

    ################################
    def setrange(self, mn, mx):
        ################################
        """
        clip the palette between mn and mx

        NB: this function should be added to get

        mn = minimum
        mx = maximum

        """

        self.t = np.arange(0, 255, 255. / (mx - mn))

        if is_scipy:
            self.r = splev(self.t, self.ar)
            self.g = splev(self.t, self.ag)
            self.b = splev(self.t, self.ab)

        self.r = np.clip(self.r, 0., 255.)
        self.g = np.clip(self.g, 0., 255.)
        self.b = np.clip(self.b, 0., 255.)

        # reconstuct the palette
        z = np.zeros(256, float)
        z[mn:mn + len(self.r)] = self.r
        self.r = z

        z = np.zeros(256, float)
        z[mn:mn + len(self.g)] = self.g
        self.g = z

        z = np.zeros(256, float)
        z[mn:mn + len(self.b)] = self.b
        self.b = z

        self.palette = b''

        for i in range(256):

            r = int(self.r[i])
            g = int(self.g[i])
            b = int(self.b[i])

            self.palette = self.palette + \
                bytearray((r,)) + bytearray((g,)) + bytearray((b,))

    ###########################

    def addFrame(self, frame):
        ###########################

        if is_tk:

            ###########################
            # frame
            ###########################

            self.Frame = frame
            self.Frame.bind('<Return>', self.revert)		  # marche pas...
            self.Frame.bind('<KeyPress-e>', self.edit) 	  # marche pas...

            ###########################
            # canvas
            ###########################

            self.Canvas = tk.Canvas(frame, height=48, width=512)
            self.Canvas.grid(column=0, row=0)
            self.Canvas.bind('<Motion>', self.motion)
            self.Canvas.bind('<B1-Motion>', self.move1)
            self.Canvas.bind('<B2-Motion>', self.move2)
            self.Canvas.bind('<B3-Motion>', self.move3)
            self.Canvas.bind('<Double-1>', self.mouse1)
            self.Canvas.bind('<Double-2>', self.edit)
            self.Canvas.bind('<Double-3>', self.mouse3)

            self.draw()

        else:
            print("addFrame : tk is not present")

    ###########################

    def change(self, name):
        ###########################

        self.name = self.check_palette_name(name)
        self.read(self.name)
        self.draw()

    ###########################
    def draw(self):
        ###########################

        if is_tk:

            palette_shape = (512, 48)

            data = chr(0)

            for i in range(0, 48):
                for j in range(0, 256):

                    data = data + chr(j)
                    data = data + chr(j)

            data = data[1:]

            # converting data
            image = Image.frombytes("P", palette_shape, data)

            # include the palette
            image.putpalette(self.palette)

            # create a Tk photo
            self.palette_pic = ImageTk.PhotoImage(image)

            # insert photo in the pannel
            if self.Canvas is not None:
                self.Canvas.create_image(
                    0., 0., anchor=tk.NW, image=self.palette_pic)

        else:

            print("draw : tk is not present")

    ###########################

    def edit(self, event):
        ###########################

        EditParams(self, self.c, self.w, self.d, self.invert)

    ###########################

    def mouse1(self, event):
        ###########################

        # find index in tables

        i = self.tables.index(self.name) + 1
        if i < len(self.tables):
            self.change(self.tables[i])

    ###########################
    def mouse2(self, event):
        ###########################

        # find index in tables

        i = self.tables.index(self.name) - 1
        if i >= 0:
            self.change(self.tables[i])

    ###########################
    def move1(self, event):
        ###########################

        self.c = self.c - (event.x - self.oevent.x)

        self.get(c=self.c, w=self.w, d=self.d)
        self.draw()

        self.oevent = event

    ###########################
    def move2(self, event):
        ###########################

        self.w = self.w + (event.x - self.oevent.x)

        self.get(c=self.c, w=self.w, d=self.d)
        self.draw()

        self.oevent = event

    ###########################
    def move3(self, event):
        ###########################

        self.d = self.d + (event.x - self.oevent.x) * (0.01 * self.d)

        if self.d < 0.:
            self.d = 1e-3
        if self.d > 256.:
            self.d = 256.

        self.get(c=self.c, w=self.w, d=self.d)
        self.draw()

        self.oevent = event

    ###########################
    def mouse3(self, event):
        ###########################

        self.c = 128
        self.w = 256
        self.d = 256.
        self.invert = 0
        self.get(c=self.c, w=self.w, d=self.d)
        self.draw()

        self.oevent = event

    ###########################
    def motion(self, event):
        ###########################

        self.oevent = event

    ###########################
    def revert(self, event):
        ###########################

        print("revert")
